#ifndef MAKEDEPEND
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdarg.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <fcntl.h>
#endif

#include "cons.h"

/* This is scratch code, not meant to be finished */

/* Copyright (C) 1992,2000 Bisqwit (http://iki.fi/bisqwit/)
 * No warranty. You are free to modify this source, and to
 * distribute modified sources, as long as you keep the
 * existing copyright messages intact and as long as you
 * remember to add your own copyright markings.
 */

unsigned LINES=0, COLS=0;

unsigned WhereX=32767, WhereY=32767;    /* Supposed to be   */
BYTE TextAttr=255;

static unsigned wx=32767, wy=32767;     /* Real coordinates */
static const int Use_VerticalTabs    = 1;
static const int Trust_DefaultColors = 0;
static int cursorvisible=1;
static BYTE OldAttr=255;

static int LastNap = EOF;

static struct termios tattr;
static struct termios saved_attributes;
#if vidcache
static BYTE *mbuf;
static BYTE *cbuf;
static BYTE *sbuf;
#endif

static BYTE set ='';
static BYTE sset='';

int AllowINT;
int AllowXON;

char *strlwr(char *string)
{
    char *temp = string;
    while(*temp)
    {
        *temp = tolower((int)*temp);
        if(*temp=='')*temp='';
        if(*temp=='')*temp='';
        if(*temp=='')*temp='';
        temp++;
    }    
    return string;
}

char *strupr(char *string)
{
    char *temp = string;
    while(*temp)
    {
        *temp = toupper((int)*temp);
        if(*temp=='')*temp='';
        if(*temp=='')*temp='';
        if(*temp=='')*temp='';
        temp++;
    }    
    return string;
}

int strnicmp(const char *eka, const char *toka, long Max)
{
    while(Max--)
    {
        char ch1 = toupper((int)*eka);
        char ch2 = toupper((int)*toka);
        if(ch1 < ch2)return -1;
        if(ch1 > ch2)return 1;
        if(!ch1)break;
        eka++;
        toka++;
    }
    return 0;
}

char *strccat(char *Dest, char Ch)
{
    register int d=strlen(Dest);
    
    Dest[d] = Ch;
    Dest[d+1] = 0;
    
    return Dest;
}

static void ClearCache(void)
{
    unsigned w = COLS*LINES;
#if vidcache
    memset(mbuf, 0, w);
    memset(cbuf, 0, w);
    memset(sbuf, 0, w);
#endif
}

void DoneCons(void)
{
    tcsetattr(STDIN_FILENO, TCSAFLUSH, &saved_attributes);
    putchar('');
}

void GetTerminalSize(void)
{
    struct winsize ws;        /* buffer for TIOCSWINSZ */
  
    if(ioctl(STDOUT_FILENO, TIOCGWINSZ, &ws) >= 0)
    {
        LINES = ws.ws_row;
        COLS  = ws.ws_col;
    }
    if(LINES==0 || COLS==0)
    {
        printf("Unknown window sizes (%d, %d), setting to 80x50\n",COLS,LINES);
        LINES=50;
        COLS=80;
    }
}

void InitCons(void)
{ 
    GetTerminalSize();
    
    fflush(stdout);
#if 0
    setvbuf(stdout, NULL, _IOFBF, 2048);
#else
    setvbuf(stdout, NULL, _IONBF, 0);
#endif

    printf("\33)0");
    
    tcgetattr(STDIN_FILENO, &saved_attributes);
    tcgetattr(STDIN_FILENO, &tattr);
    tattr.c_lflag &= ~(ICANON|ECHO);
    if(!AllowINT)tattr.c_lflag &= ~(ISIG);                /* enable ^C */
    if(!AllowXON)tattr.c_iflag &= ~(IXON);                /* enable ^Q */
    tattr.c_cc[VMIN] = 1;
    tattr.c_cc[VTIME] = 0;
    tcsetattr(STDIN_FILENO, TCSAFLUSH, &tattr);
    fcntl(STDIN_FILENO,F_SETFL,fcntl(STDIN_FILENO,F_GETFL)|O_NONBLOCK);
    
#if vidcache
    mbuf = (BYTE *)malloc(COLS * LINES);
    cbuf = (BYTE *)malloc(COLS * LINES);
    sbuf = (BYTE *)malloc(COLS * LINES);
#endif
    ClearCache();
}

void Error(const char *Where)
{
     printf("\33[0;1;31mProg: %s() error %d `\33[37m%s\33[31m'\33[m\n\n", Where, errno, sys_errlist[errno]);
    exit(errno);
}

void GotoXY(unsigned x, unsigned y)
{
    WhereX=x, WhereY=y;
}

static void FlushXY(void)
{
    register unsigned x=WhereX, y=WhereY;
    register unsigned oldx=wx, oldy=wy;
    if(x==oldx && y==oldy)Done:return;

    wx=x--, wy=y--;
    oldx--, oldy--;
    
    if(sset != '')putchar(sset='');
       
    if(!x)
    {
        if(y==oldy+4) { printf("\n\n\n\n"); goto Done; }
        if(y==oldy+3) { printf("\n\n\n"); goto Done; }
        if(y==oldy+2) { printf("\n\n"); goto Done; }
        if(y==oldy+1) { putchar('\n'); goto Done; }
        if(!y) { printf("\33[H"); goto Done; }
        if(oldx)putchar('\r');
        goto FixUp;
    }
    if(x==oldx)goto FixUp;
    if(x==oldx-1)
    {
        putchar('\b');
    FixUp:   if(y == oldy-1)printf("\33[A");
        else if(Use_VerticalTabs && y == oldy+1)putchar('\v');
        else if(Use_VerticalTabs && y == oldy+2)printf("\v\v");
        else if(Use_VerticalTabs && y == oldy+3)printf("\v\v\v");
        else if(y < oldy)printf("\33[%dA", oldy-y);
        else if(y > oldy)printf("\33[%dB", y-oldy);
        goto Done;
    }
    if(y == oldy)
    {
        /***/if(x == oldx+1)printf("\33[C");
        else if(x == oldx-2)printf("\b\b");
        else if(x == oldx-3)printf("\b\b\b");
        else if(x < oldx)printf("\33[%dD", oldx-x);
        else if(x > oldx)printf("\33[%dC", x-oldx);
        goto Done;
    }
    printf("\33[%d;%dH", wy, wx);
}

void SetAttr(BYTE n)
{
    TextAttr = n;
}

void FlushAttr(void)
{
    if(OldAttr != TextAttr)
    {
        if(sset != '')putchar(sset='');
        printf("\33[");
        if(TextAttr != 7 || !Trust_DefaultColors)
        {
            static const char Swap[8] = "04261537";

            int pp=0, r=0;
       
            if((OldAttr&0x80) > (TextAttr&0x80)
            || (OldAttr&0x08) > (TextAttr&0x08))
            {
                putchar('0'); pp = 1;
                OldAttr = 0x07;
                r = 1; /* !Trust_DefaultColors; */
            }
            
            if((TextAttr&0x08) && !(OldAttr&0x08)){if(pp)putchar(';');putchar('1');pp=1;}
            if((TextAttr&0x80) && !(OldAttr&0x80)){if(pp)putchar(';');putchar('5');pp=1;}
               
            if((TextAttr&7) != (OldAttr&7) || r)
            {
                if(pp)putchar(';');
                printf("3%c", Swap[TextAttr&7]);
                pp = 1;
            }
            if((TextAttr&0x70) != (OldAttr&0x70) || r)
            {
                if(pp)putchar(';');
                printf("4%c", Swap[(TextAttr>>4)&7]);
            }
        }
        putchar('m');
        OldAttr = TextAttr;
    }
}

void ShowCursor(void)
{
    if(!cursorvisible)
    {
        FlushAttr();
        FlushXY();
        printf("\33[?25h");
        cursorvisible=1;
    }
}
void HideCursor(void)
{
    if(cursorvisible)
    {
        cursorvisible=0;
        printf("\33[?25l");
    }
}

void ClrScr(void)
{
#if vidcache
    memset(mbuf, ' ',      COLS*LINES);
    memset(cbuf, TextAttr, COLS*LINES);
    memset(sbuf, '',      COLS*LINES);
#endif
    if(sset != '')putchar(sset='');
    FlushAttr();
    printf("\33[2J");
    GotoXY(1, 1);
}
void RefreshScreen(void)
{
    unsigned x, y;
    unsigned w;
    BYTE *Mbuf, *Cbuf, *Sbuf, Set, ta;
    
    GetTerminalSize();
    
    w = COLS*LINES;
    Mbuf = (BYTE *)malloc(w);
    Cbuf = (BYTE *)malloc(w);
    Sbuf = (BYTE *)malloc(w);
    Set = set;
    ta = TextAttr;
    
#if vidcache
    memcpy(Mbuf, mbuf, w);
    memcpy(Cbuf, cbuf, w);
    memcpy(Sbuf, sbuf, w);
#endif
    
    SetAttr(0x07);
    ClrScr();
    w=0;
    for(y=0; y<LINES; y++)
    {
        GotoXY(1, y+1);
        for(x=0; x<COLS; x++)
        {
            set = Sbuf[w];
            SetAttr(Cbuf[w]);
            Gputchar(Mbuf[w++]);
        }
    }
    set = Set;
    SetAttr(ta);
    
    free(Mbuf);
    free(Cbuf);
    free(Sbuf);
}
void ClrEOL(void)
{
    FlushAttr();
    FlushXY();
    printf("\33[K");
#if vidcache
    memset(mbuf+(WhereY-1)*COLS+WhereX-1, ' ',      COLS-(WhereX-1));
    memset(cbuf+(WhereY-1)*COLS+WhereX-1, TextAttr, COLS-(WhereX-1));
    memset(sbuf+(WhereY-1)*COLS+WhereX-1, '',      COLS-(WhereX-1));
#endif
}

BYTE Gputchar(BYTE c)
{
    unsigned w;
    if(c < 32)
        return set=c;
    
    #if vidcache
    w = (WhereY-1)*COLS + WhereX-1;
    if(mbuf[w] == c && cbuf[w] == TextAttr)
    {
        WhereX++;
        return c;
    }
    cbuf[w] = TextAttr;
    #endif
    FlushXY();
    FlushAttr();
    WhereX++, wx++;
    #if vidcache
    if((sbuf[w]=set) != sset)putchar(sset=set);
    return putchar(mbuf[w]=c);
    #else
    return putchar(c);
    #endif
}
int Gprintf(const char *fmt, ...)
{
    int n;
    va_list ap;
    char Buf[2048], *s;
    va_start(ap, fmt);
    vsprintf(Buf, fmt, ap);
    va_end(ap);
    for(n=0, s=Buf; *s; ++n)
    {
    	BYTE c = *s++;
        unsigned w;
        if(c < 32)
            set=c;
        else
        {
            #if vidcache
            
            w = (WhereY-1)*COLS + WhereX-1;
            if(mbuf[w] == c && cbuf[w] == TextAttr
            && (s[1] == mbuf[w+1]))
                WhereX++;
            else
            {
                cbuf[w] = TextAttr;
                FlushXY();
                FlushAttr();
                WhereX++, wx++;
                if((sbuf[w] = set) != sset)
                    putchar(sset = set);
                
                putchar(mbuf[w] = c);
            }
            
            #else
            
            FlushXY();
            FlushAttr();
            WhereX++, wx++;
            putchar(c);
            
            #endif
        }
    }
    return n;
}

int getch(void)
{
    int a, temp;
    if(LastNap != EOF)
    {
        temp = LastNap;
        LastNap = EOF;
        return temp;
    }
    a = fcntl(STDIN_FILENO, F_GETFL);
    fcntl(STDIN_FILENO, F_SETFL, a & ~O_NONBLOCK);
    
    temp = fgetc(stdin);
    
    fcntl(STDIN_FILENO, F_SETFL, a);
    
    return temp;
}

int kbhit(void) 
{
    if(cursorvisible)FlushXY();
#if 0
    fflush(stdout);
#endif
    
    if(LastNap != EOF)return 1;
    LastNap = fgetc(stdin);
    return LastNap != EOF;
}

int getkey(void)
{
    /* EXSD = PNBF = ABDC */
    int b;
    
    kbhit();

    b = getch();
    if(b==Ctrl+'P')return KbUp;
    if(b==Ctrl+'N')return KbDn;
/*  if(b==Ctrl+'B')return KbLt; */
    if(b==Ctrl+'F')return KbRt;
    if(b==Ctrl+'A')return Home;
    if(b==Ctrl+'E')return End;
    if(b==Ctrl+'U')return PgUp;
    if(b==Ctrl+'V')return PgDn;
    if(b==Ctrl+'I')return Tab;
    if(b==Ctrl+'O')return ShiftTab;
    if(b==Ctrl+'H')return Ctrl+Backspace;
    if(b==127     )return Backspace;
    if(b==Ctrl+'D')return Del;
    if(b==Ctrl+'_')return Ins;
    if(b==Esc)
    {
        int zzz = 0;
        while(!kbhit())
        {
            if(++zzz > 20)return Esc;    /* 200ms timeout */
            usleep(10000);
        }
        switch(b=getch())
        {
            case '[':
            {
                int c;
                for(b=c=0;b!='['&&b!='~'&&(b<'a'||b>'z')&&(b<'A'||b>'Z');b=getch())
                    if(b>='0'&&b<='9')c=c*10+b-'0';
                switch(b)
                {
                    case 'A': return KbUp;
                    case 'B': return KbDn;
                    case 'D': return KbLt;
                    case 'C': return KbRt;
                    case 'P': return Pause;
                    case '[':
                        switch(getch())
                        {
                            case 'A': return F1;
                            case 'B': return F2;
                            case 'C': return F3;
                            case 'D': return F4;
                            case 'E': return F5;
                        }
                        break;
                    case '~':
                        switch(c)
                        {
                            case 2: return Ins;
                            case 3: return Del;
                            case 5: return PgUp;
                            case 6: return PgDn;
                            case 17: return F6;
                            case 18: return F7;
                            case 19: return F8;
                            case 20: return F9;
                            case 21: return F10;
                        }
                }
                break;
            }
            default:
                return Alt + toupper(b);
        }
    }
    return b;
}
