#ifndef pokemud_mididev_h
#define pokemud_mididev_h

#include <cstdio>
#include <unistd.h>
#include <fcntl.h>

#include "config.h"

#if defined(TIMIDITYHOST) && defined(TIMIDITYPORT)
#include "mudconn.h"
#include <string>
#endif

class mididev
{
    int fd;
    bool timeractive;
    
    #if defined(TIMIDITYHOST) && defined(TIMIDITYPORT)
    mudconn timidityconn[2];
    #endif
    
    void write(unsigned char c)
    {
        char buf[4] = {5,c,1,0};
        std::write(fd, buf, 4);
    }
    void timercommand(unsigned char c, unsigned parm)
    {
    	char buf[4] = {0x81, c, 0, 0};
    	std::write(fd, buf, 4);
    	std::write(fd, &parm, 4);
    }
public:
    mididev();
    ~mididev()
    {
        std::close(fd);
    }
    int getfd() const
    {
        return fd;
    }
    mididev &operator << (unsigned char c) { write(c); return *this; }
    void noteoff(int ch, unsigned char note, unsigned char pressure=0)
    {
        *this << 0x80+ch << note << pressure;
    }
    void noteon(int ch, unsigned char note, unsigned char pressure=127)
    {
        *this << 0x90+ch << note << pressure;
    }
    void ctrl(int ch, unsigned char ctrl, unsigned char value)
    {
        *this << 0xB0+ch << ctrl << value;
    }
    void patch(int ch, unsigned char ins)
    {
        *this << 0xC0+ch << ins-1;
    }
    void sleep(unsigned us)
    {
    	if(timeractive)timercommand(2, us);  // abs wait
    	timercommand(4, 0);   // timer start
    	timeractive = true;
    }
};

#endif
