#include <sys/time.h>

#include "terminal.h"
#include "config.h"

#if !SIMPLE
// From socket:
void terminal::input()
{
    char Buf[512], *s;
    int n = std::read(fd, Buf, 512);
    if(n <= 0)
    {
        if(!n && forgive) { --forgive; return; }
        writefd("\n\nConnection closed by foreign host.\n", 1);
        dead=true;
        return;
    }
    for(s=Buf; n; --n)
    {
        char c = *s++;
        switch(inputstate)
        {
            case midi:
                if(c == 10)
                {
                    musicdev->upload(midistr);
                    inputstate = normal;
                }
                else if(c != 13)midistr += char(c);
                break;
            case normal:
                if(c == 27) // esc
                    inputstate = esc;
                else
                    anyout(c);
                break;
            case esc:
                if(c == '[')
                {
                    inputstate = data;
                    iansiparm[iansipos = iansihit = 0] = 0;
                }
                else
                {
                    anyout('\33');
                    anyout(c);
                }
                break;
            case data:
                if(c == ';')
                {
                    if((unsigned)iansipos < (sizeof iansiparm) / (sizeof(iansiparm[0]))-1)
                        iansiparm[++iansipos] = 0;
                    iansihit = 0;
                }
                else if(c >= '0' && c <= '9')
                {
                    iansiparm[iansipos] = iansiparm[iansipos]*10 + c - '0';
                    iansihit = 1;
                }
                else
                {
                    if(iansihit)++iansipos;
                    inputstate = normal;
                    switch(c)
                    {
                        // No need to translate, as we don't use it.
                        // Yet.
                        // Should use, in the prompt colouring.
                        case 'm':
                        {
                            int newattr = textattr;
                            for(int a=0; a<iansipos; ++a)
                            {
                                switch(iansiparm[a])
                                {
                                    case 0: newattr = 7; break;
                                    case 1: newattr |= 8; break;
                                    case 21:
                                    case 22: newattr &= ~8; break;
                                    case 5: newattr |= 0x80; break;
                                    case 25: newattr &= ~0x80; break;
                                    case 7:
                                    case 27: newattr ^= 0x77; break;
                                    case 30: newattr = (newattr&0xF8); break;
                                    case 31: newattr = (newattr&0xF8)|4; break;
                                    case 32: newattr = (newattr&0xF8)|2; break;
                                    case 33: newattr = (newattr&0xF8)|6; break;
                                    case 34: newattr = (newattr&0xF8)|1; break;
                                    case 35: newattr = (newattr&0xF8)|5; break;
                                    case 36: newattr = (newattr&0xF8)|3; break;
                                    case 37: newattr = (newattr&0xF8)|7; break;
                                    case 40: newattr = (newattr&0x8F); break;
                                    case 41: newattr = (newattr&0x8F)|0x40; break;
                                    case 42: newattr = (newattr&0x8F)|0x20; break;
                                    case 43: newattr = (newattr&0x8F)|0x60; break;
                                    case 44: newattr = (newattr&0x8F)|0x10; break;
                                    case 45: newattr = (newattr&0x8F)|0x50; break;
                                    case 46: newattr = (newattr&0x8F)|0x30; break;
                                    case 47: newattr = (newattr&0x8F)|0x70; break;
                                }
                            }
                            textattr = newattr;
                   setattr(textattr, true);
                            break;
                        }
                        case 'c':
                            switch(iansiparm[0])
                            {
                                case 0: writefd("\33[[?6c"); break;
                                case 1:
                                    hideoutput();
                                    prompt = inputline;
                                    inputline = "";
                                    inputpos = 0;
                                    showoutput();
                                    break;
                                case 2:
                                    inputstate = midi;
                                    midistr = "";
                                    break;
                            }
                            break;
                        case 'n':
                            switch(iansiparm[0])
                            {
                                case 6:
                                    // cursor report... blah.
                                    writefd("\33[0;0R");
                                    break;
                            }
                            break;
                        default:
                            ansiout(c);
                    }
                }
        }
    }
}
#endif

// From keyboard:
void terminal::anykey(unsigned char c)
{
    switch(c)
    {
        case '':
            fixoutput(0, outputline);
            break;
        case '':
            fixoutput(outputline.size(), outputline);
            break;
        case '':
        case '':
        {
            unsigned n = outputpos;
            string tmp = outputline;
            if(n > 0)--n;
            tmp.erase(n, 1);
            fixoutput(n, tmp);
            break;
        }
        case '':
        {
            string tmp = outputline;
            tmp.erase(outputpos, 1);
            fixoutput(outputpos, tmp);
            break;
        }
        case '':
            if(outputpos>0)fixoutput(outputpos-1, outputline);
            break;
        case '':
            if(outputpos<outputline.size())fixoutput(outputline.size()+1, outputline);
            break;
        case '':
        	break;
        case '':
        	break;
        case '':
            fixoutput(0, "");
            break;
        case '\r':
        case '\n':
            writefd("\r\n", 1);
            
            if(fd < 0)
            {
	        	if(outputline == "q")
	        		dead = true;
	        	else
	    	        musicdev->upload(outputline);
    	    }
    	    else
    	    {
	            if(outputline.size() && outputline[0] == '/')
    	            musicdev->upload(outputline.substr(1));
        	    else
            	{
    	            outputline += '\r';
	                writefd(outputline);
        	    }
        	}
            outputline = "";
            outputpos = 0;
            outputvisible = false;
            justgavedata = true;
            break;
        default:
            if((c >= ' ' && c < 127) || (c > 160 && c < 250))
            {
                string tmp = outputline;
                unsigned pos = outputpos;
                tmp.insert(pos++, 1, c);
                fixoutput(pos, tmp);
            }
    }
}

void terminal::run(int fdee, music &midi)
{
    fd = fdee;
    
    musicdev = &midi;
    
    inputpos = 0;
    inputline = "";
    inputstate = normal;
    outputpos = 0;
    outputline = "";
    outputstate = normal;
    outputvisible = false;
    forgive = 4;
    justgavedata = false;
    currentattr = 0xFF;
    textattr = 7;
    if(fd >= 0)
		writefd("\33[[?6c\r");
	else
	    prompt = "Enter command: ";
    
    unsigned idle=0;
    while(!dead)
    {
        int val;
        struct timeval tv = {0,10000};
        do {
        	#ifndef linux
        	/* can't trust the values :( */
        	tv.tv_sec = 0;
        	tv.tv_usec = 10000;
        	#endif
            fd_set fds;
            FD_ZERO(&fds);
            FD_SET(0, &fds);
            if(fd < 0)
	            val = select(1, &fds,NULL,NULL, &tv);
	        else
	        {
	            FD_SET(fd, &fds);
	            val = select(fd+1, &fds,NULL,NULL, &tv);
	        }
            if(val < 0)return;
            if(val > 0)
            {
                idle = 0;
                if(FD_ISSET(0, &fds)) output();
#if !SIMPLE
                if(fd >= 0)
	                if(FD_ISSET(fd, &fds))input();
#endif
            }
            if(!justgavedata || ++idle > 50)
            {
                showoutput();
                justgavedata = false;
            }
            // Tm ehto... tm looppi uudelleen kytt tv:n
            // arvoa, ja siksi toimii vain Linuxissa (sek ehk
            // joissain muissa System V -jrjestelmiss.
            // Tmn jrjestelyn avulla saadaan tuo
            // midi.tick() tikittmn oikeaan tahtiin
            // vaikka linjalta tuuppaisikin tavaraa
            // vliin lyhentmn viivett.
        } while(val && !dead);
        
        // 7 = mutuluku
        for(int i=0; i<7; ++i)
            midi.tick();
    }
}
